
(********************************************************************
  module Slit
    generic Slit with Dirty Flag
 ********************************************************************)

module type Instance =
  sig
    type item
    type label
  end

module Make = functor (M : Instance) ->
  struct
    open ListZipper

    (* types *)
    type tree = Leaf of M.item
              | Node of M.label * tree list
    type slit = (tree list_zipper * bool) list

    (* move operations 1 *)
    let prev = function (hd,m)::tl -> (lz_prev hd, m)::tl
    let next = function (hd,m)::tl -> (lz_next hd, m)::tl

    (* move operations 2 *)
    let down = function (hd,m)::tl ->
      match (lz_get_elem hd) with Node(_, children) ->
        (lz_of_list children,false)::(hd,m)::tl

    let up = function (hd1,d1)::(hd2,d2)::tl ->
      if d1 then
        match (lz_get_elem hd2) with Node(label, _) ->
          (lz_set_elem (Node(label, list_of_lz hd1)) hd2, true)::tl
      else
        (hd2,d2)::tl

    (* edit operations *)
    let get_elem        = function (hd,m)::tl -> lz_get_elem hd
    let set_elem e      = function (hd,m)::tl -> (lz_set_elem e      hd,true)::tl
    let insert_after  e = function (hd,m)::tl -> (lz_insert_after e  hd,true)::tl
    let insert_before e = function (hd,m)::tl -> (lz_insert_before e hd,true)::tl
    let remove_elem     = function (hd,m)::tl -> (lz_remove_elem     hd,true)::tl

    (* conversion *)
    let slit_of_tree tr = [lz_of_list [tr], false]
    let tree_of_slit sl =
      let rec iter f sl = try iter f (f sl) with _ -> sl in
        get_elem (iter prev (iter up sl))
  end
